@@ -99,6 +99,40 @@ static void decodeIndexV1()
 	assert(memcmp(decoded, kIndexBufferTricky, sizeof(kIndexBufferTricky)) == 0);
 }
 
+static void decodeIndexV1More()
+{
+	const unsigned char input[] = {
+	    0xe1, 0xf0, 0x10, 0xfe, 0xff, 0xf0, 0x0c, 0xff, 0x02, 0x02, 0x02, 0x00, 0x76, 0x87, 0x56, 0x67,
+	    0x78, 0xa9, 0x86, 0x65, 0x89, 0x68, 0x98, 0x01, 0x69, 0x00, 0x00, // clang-format
+	};
+
+	const unsigned int ib[] = {0, 1, 2, 2, 1, 3, 4, 6, 5, 7, 8, 9};
+	const size_t index_count = sizeof(ib) / sizeof(ib[0]);
+
+	std::vector<unsigned char> buffer(input, input + sizeof(input));
+
+	unsigned int decoded[index_count];
+	assert(meshopt_decodeIndexBuffer(decoded, index_count, 4, &buffer[0], buffer.size()) == 0);
+	assert(memcmp(decoded, ib, sizeof(ib)) == 0);
+}
+
+static void decodeIndexV1ThreeEdges()
+{
+	const unsigned char input[] = {
+	    0xe1, 0xf0, 0x20, 0x30, 0x40, 0x00, 0x76, 0x87, 0x56, 0x67, 0x78, 0xa9, 0x86, 0x65, 0x89, 0x68,
+	    0x98, 0x01, 0x69, 0x00, 0x00, // clang-format
+	};
+
+	const unsigned int ib[] = {0, 1, 2, 1, 0, 3, 2, 1, 4, 0, 2, 5};
+	const size_t index_count = sizeof(ib) / sizeof(ib[0]);
+
+	std::vector<unsigned char> buffer(input, input + sizeof(input));
+
+	unsigned int decoded[index_count];
+	assert(meshopt_decodeIndexBuffer(decoded, index_count, 4, &buffer[0], buffer.size()) == 0);
+	assert(memcmp(decoded, ib, sizeof(ib)) == 0);
+}
+
 static void decodeIndex16()
 {
 	const size_t index_count = sizeof(kIndexBuffer) / sizeof(kIndexBuffer[0]);
@@ -377,6 +411,50 @@ static void decodeVertexV0()
 	assert(memcmp(decoded, kVertexBuffer, sizeof(kVertexBuffer)) == 0);
 }
 
+static void decodeVertexV0More()
+{
+	const unsigned char expected[] = {
+	    0, 0, 0, 0, 0, 1, 2, 8, 0, 2, 4, 16, 0, 3, 6, 24,
+	    0, 4, 8, 32, 0, 5, 10, 40, 0, 6, 12, 48, 0, 7, 14, 56,
+	    0, 8, 16, 64, 0, 9, 18, 72, 0, 10, 20, 80, 0, 11, 22, 88,
+	    0, 12, 24, 96, 0, 13, 26, 104, 0, 14, 28, 112, 0, 15, 30, 120, // clang-format :-/
+	};
+
+	const unsigned char input[] = {
+	    0xa0, 0x00, 0x01, 0x2a, 0xaa, 0xaa, 0xaa, 0x02, 0x04, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x44,
+	    0x03, 0x00, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
+	    0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+	    0x00, // clang-format :-/
+	};
+
+	unsigned char decoded[sizeof(expected)];
+	assert(meshopt_decodeVertexBuffer(decoded, 16, 4, input, sizeof(input)) == 0);
+	assert(memcmp(decoded, expected, sizeof(expected)) == 0);
+}
+
+static void decodeVertexV0Mode2()
+{
+	const unsigned char expected[] = {
+	    0, 0, 0, 0, 4, 5, 6, 7, 8, 10, 12, 14, 12, 15, 18, 21,
+	    16, 20, 24, 28, 20, 25, 30, 35, 24, 30, 36, 42, 28, 35, 42, 49,
+	    32, 40, 48, 56, 36, 45, 54, 63, 40, 50, 60, 70, 44, 55, 66, 77,
+	    48, 60, 72, 84, 52, 65, 78, 91, 56, 70, 84, 98, 60, 75, 90, 105, // clang-format :-/
+	};
+
+	const unsigned char input[] = {
+	    0xa0, 0x02, 0x08, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x88, 0x02, 0x0a, 0xaa, 0xaa, 0xaa, 0xaa,
+	    0xaa, 0xaa, 0xaa, 0x02, 0x0c, 0xcc, 0xcc, 0xcc, 0xcc, 0xcc, 0xcc, 0xcc, 0x02, 0x0e, 0xee, 0xee,
+	    0xee, 0xee, 0xee, 0xee, 0xee, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+	    0x00, 0x00, 0x00, 0x00, 0x00, // clang-format :-/
+	};
+
+	unsigned char decoded[sizeof(expected)];
+	assert(meshopt_decodeVertexBuffer(decoded, 16, 4, input, sizeof(input)) == 0);
+	assert(memcmp(decoded, expected, sizeof(expected)) == 0);
+}
+
 static void decodeVertexV1()
 {
 	const size_t vertex_count = sizeof(kVertexBuffer) / sizeof(kVertexBuffer[0]);
@@ -538,7 +616,7 @@ static void decodeVertexDeltas()
 	}
 
 	std::vector<unsigned char> buffer(meshopt_encodeVertexBufferBound(16, 8));
-	buffer.resize(meshopt_encodeVertexBufferLevel(&buffer[0], buffer.size(), data, 16, 8, 2));
+	buffer.resize(meshopt_encodeVertexBufferLevel(&buffer[0], buffer.size(), data, 16, 8, 2, -1));
 
 	unsigned short decoded[16 * 4];
 	assert(meshopt_decodeVertexBuffer(decoded, 16, 8, &buffer[0], buffer.size()) == 0);
@@ -559,7 +637,7 @@ static void decodeVertexBitXor()
 	}
 
 	std::vector<unsigned char> buffer(meshopt_encodeVertexBufferBound(16, 16));
-	buffer.resize(meshopt_encodeVertexBufferLevel(&buffer[0], buffer.size(), data, 16, 16, 3));
+	buffer.resize(meshopt_encodeVertexBufferLevel(&buffer[0], buffer.size(), data, 16, 16, 3, -1));
 
 	unsigned int decoded[16 * 4];
 	assert(meshopt_decodeVertexBuffer(decoded, 16, 16, &buffer[0], buffer.size()) == 0);
@@ -1068,18 +1146,12 @@ static void sphereBounds()
 	    1, 0, 1, 3, // clang-format
 	};
 
-	// without the radius, the center is somewhere inside the tetrahedron
-	// note that we currently compute a somewhat suboptimal sphere here due to the tetrahedron being perfecly axis aligned
+	// without the radius, the center is inside the tetrahedron
 	meshopt_Bounds bounds = meshopt_computeSphereBounds(vbr, 4, sizeof(float) * 4, NULL, 0);
-	assert(bounds.radius < 0.97f);
-
-	// forcing a better initial guess for the center by using different radii produces a close to optimal sphere
-	float eps[4] = {1e-3f, 2e-3f, 3e-3f, 4e-3f};
-	meshopt_Bounds boundse = meshopt_computeSphereBounds(vbr, 4, sizeof(float) * 4, eps, sizeof(float));
-	assert(fabsf(boundse.center[0] - 0.5f) < 1e-2f);
-	assert(fabsf(boundse.center[1] - 0.5f) < 1e-2f);
-	assert(fabsf(boundse.center[2] - 0.5f) < 1e-2f);
-	assert(boundse.radius < 0.87f);
+	assert(fabsf(bounds.center[0] - 0.5f) < 1e-2f);
+	assert(fabsf(bounds.center[1] - 0.5f) < 1e-2f);
+	assert(fabsf(bounds.center[2] - 0.5f) < 1e-2f);
+	assert(bounds.radius < 0.87f);
 
 	// when using the radius, the last sphere envelops the entire set
 	meshopt_Bounds boundsr = meshopt_computeSphereBounds(vbr, 4, sizeof(float) * 4, vbr + 3, sizeof(float) * 4);
@@ -1249,6 +1321,73 @@ static void meshletsMax()
 	}
 }
 
+static void meshletsSpatial()
+{
+	// two tetrahedrons far apart
+	float vb[2 * 4 * 3] = {
+	    0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1,
+	    10, 0, 0, 11, 0, 0, 10, 1, 0, 10, 0, 1, // clang-format :-/
+	};
+
+	unsigned int ib[2 * 4 * 3] = {
+	    0, 1, 2, 0, 2, 3, 0, 3, 1, 1, 3, 2,
+	    4, 5, 6, 4, 6, 7, 4, 7, 5, 5, 7, 6, // clang-format :-/
+	};
+
+	// up to 2 meshlets with min_triangles=4
+	assert(meshopt_buildMeshletsBound(2 * 4 * 3, 16, 4) == 2);
+
+	meshopt_Meshlet ml[2];
+	unsigned int mv[2 * 16];
+	unsigned char mt[2 * 8 * 3]; // 2 meshlets with up to 8 triangles
+
+	// with strict limits, we should get one meshlet (maxt=8) or two (maxt=4)
+	assert(meshopt_buildMeshletsSpatial(ml, mv, mt, ib, sizeof(ib) / sizeof(ib[0]), vb, 8, sizeof(float) * 3, 16, 8, 8, 0.f) == 1);
+	assert(ml[0].triangle_count == 8);
+	assert(ml[0].vertex_count == 8);
+
+	assert(meshopt_buildMeshletsSpatial(ml, mv, mt, ib, sizeof(ib) / sizeof(ib[0]), vb, 8, sizeof(float) * 3, 16, 4, 4, 0.f) == 2);
+	assert(ml[0].triangle_count == 4);
+	assert(ml[0].vertex_count == 4);
+	assert(ml[1].triangle_count == 4);
+	assert(ml[1].vertex_count == 4);
+
+	// with maxv=4 we should get two meshlets since we can't accomodate both
+	assert(meshopt_buildMeshletsSpatial(ml, mv, mt, ib, sizeof(ib) / sizeof(ib[0]), vb, 8, sizeof(float) * 3, 4, 4, 8, 0.f) == 2);
+	assert(ml[0].triangle_count == 4);
+	assert(ml[0].vertex_count == 4);
+	assert(ml[1].triangle_count == 4);
+	assert(ml[1].vertex_count == 4);
+}
+
+static void meshletsSpatialDeep()
+{
+	const int N = 400;
+	const size_t max_vertices = 4;
+	const size_t max_triangles = 4;
+
+	float vb[(N + 1) * 3];
+	unsigned int ib[N * 3];
+
+	vb[0] = vb[1] = vb[2] = 0;
+
+	for (size_t i = 0; i < N; ++i)
+	{
+		vb[(i + 1) * 3 + 0] = vb[(i + 1) * 3 + 1] = vb[(i + 1) * 3 + 2] = powf(1.2f, float(i));
+
+		ib[i * 3 + 0] = 0;
+		ib[i * 3 + 1] = ib[i * 3 + 2] = unsigned(i + 1);
+	}
+
+	size_t max_meshlets = meshopt_buildMeshletsBound(N * 3, max_vertices, max_triangles);
+	std::vector<meshopt_Meshlet> meshlets(max_meshlets);
+	std::vector<unsigned int> meshlet_vertices(max_meshlets * max_vertices);
+	std::vector<unsigned char> meshlet_triangles(max_meshlets * max_triangles * 3);
+
+	size_t result = meshopt_buildMeshletsSpatial(&meshlets[0], &meshlet_vertices[0], &meshlet_triangles[0], &ib[0], N * 3, &vb[0], N + 1, sizeof(float) * 3, max_vertices, max_triangles, max_triangles, 0.f);
+	assert(result == N);
+}
+
 static void partitionBasic()
 {
 	// 0   1   2
@@ -1257,25 +1396,94 @@ static void partitionBasic()
 	//     9
 	// 10 11  12
 	const unsigned int ci[] = {
-	    0, 1, 3, 4, 5, 6,   //
-	    1, 2, 3, 6, 7, 8,   //
-	    4, 5, 6, 9, 10, 11, //
-	    6, 7, 8, 9, 11, 12, //
+	    0, 1, 3, 4, 5, 6,
+	    1, 2, 3, 6, 7, 8,
+	    4, 5, 6, 9, 10, 11,
+	    6, 7, 8, 9, 11, 12, // clang-format :-/
 	};
 
 	const unsigned int cc[4] = {6, 6, 6, 6};
 	unsigned int part[4];
 
-	assert(meshopt_partitionClusters(part, ci, sizeof(ci) / sizeof(ci[0]), cc, 4, 13, 1) == 4);
+	assert(meshopt_partitionClusters(part, ci, sizeof(ci) / sizeof(ci[0]), cc, 4, NULL, 13, 0, 1) == 4);
 	assert(part[0] == 0 && part[1] == 1 && part[2] == 2 && part[3] == 3);
 
-	assert(meshopt_partitionClusters(part, ci, sizeof(ci) / sizeof(ci[0]), cc, 4, 13, 2) == 2);
+	assert(meshopt_partitionClusters(part, ci, sizeof(ci) / sizeof(ci[0]), cc, 4, NULL, 13, 0, 2) == 2);
 	assert(part[0] == 0 && part[1] == 0 && part[2] == 1 && part[3] == 1);
 
-	assert(meshopt_partitionClusters(part, ci, sizeof(ci) / sizeof(ci[0]), cc, 4, 13, 4) == 1);
+	assert(meshopt_partitionClusters(part, ci, sizeof(ci) / sizeof(ci[0]), cc, 4, NULL, 13, 0, 4) == 1);
 	assert(part[0] == 0 && part[1] == 0 && part[2] == 0 && part[3] == 0);
 }
 
+static void partitionSpatial()
+{
+	const unsigned int ci[] = {
+	    0, 1, 2,
+	    0, 3, 4,
+	    0, 5, 6, // clang-format :-/
+	};
+
+	const float vb[] = {
+	    0, 0, 0,
+	    1, 0, 0, 0, 1, 0,
+	    0, 2, 0, 2, 0, 0,
+	    -1, 0, 0, 0, -1, 0, // clang-format :-/
+	};
+
+	const unsigned int cc[3] = {3, 3, 3};
+	unsigned int part[3];
+
+	assert(meshopt_partitionClusters(part, ci, sizeof(ci) / sizeof(ci[0]), cc, 3, NULL, 7, 0, 2) == 2);
+	assert(part[0] == 0 && part[1] == 0 && part[2] == 1);
+
+	assert(meshopt_partitionClusters(part, ci, sizeof(ci) / sizeof(ci[0]), cc, 3, vb, 7, sizeof(float) * 3, 2) == 2);
+	assert(part[0] == 0 && part[1] == 1 && part[2] == 0);
+}
+
+static int remapCustomFalse(void*, unsigned int, unsigned int)
+{
+	return 0;
+}
+
+static int remapCustomTrue(void*, unsigned int, unsigned int)
+{
+	return 1;
+}
+
+static void remapCustom()
+{
+	const float vb[] = {
+	    0, 0, 0,
+	    1, 0, 0,
+	    0, 1, 0,
+	    0, 0, 1,
+	    1, 0, 0,
+	    0, -0.f, 1, // clang-format
+	};
+
+	unsigned int remap[6];
+	size_t res;
+
+	res = meshopt_generateVertexRemapCustom(remap, NULL, 6, vb, 6, sizeof(float) * 3, NULL, NULL);
+	assert(res == 4);
+	for (int i = 0; i < 4; ++i)
+		assert(remap[i] == unsigned(i));
+	assert(remap[4] == 1);
+	assert(remap[5] == 3);
+
+	res = meshopt_generateVertexRemapCustom(remap, NULL, 6, vb, 6, sizeof(float) * 3, remapCustomTrue, NULL);
+	assert(res == 4);
+	for (int i = 0; i < 4; ++i)
+		assert(remap[i] == unsigned(i));
+	assert(remap[4] == 1);
+	assert(remap[5] == 3);
+
+	res = meshopt_generateVertexRemapCustom(remap, NULL, 6, vb, 6, sizeof(float) * 3, remapCustomFalse, NULL);
+	assert(res == 6);
+	for (int i = 0; i < 6; ++i)
+		assert(remap[i] == unsigned(i));
+}
+
 static size_t allocCount;
 static size_t freeCount;
 
@@ -2081,6 +2289,36 @@ static void simplifyPruneCleanup()
 	assert(memcmp(ib, expected, sizeof(expected)) == 0);
 }
 
+static void simplifyPruneFunc()
+{
+	unsigned int ib[] = {
+	    0, 1, 2,
+	    3, 4, 5,
+	    6, 7, 8, // clang-format :-/
+	};
+
+	float vb[] = {
+	    0, 0, 0,
+	    0, 1, 0,
+	    1, 0, 0,
+	    0, 0, 1,
+	    0, 2, 1,
+	    2, 0, 1,
+	    0, 0, 2,
+	    0, 4, 2,
+	    4, 0, 2, // clang-format :-/
+	};
+
+	unsigned int expected[] = {
+	    6,
+	    7,
+	    8,
+	};
+
+	assert(meshopt_simplifyPrune(ib, ib, 9, vb, 9, 12, 0.5f) == 3);
+	assert(memcmp(ib, expected, sizeof(expected)) == 0);
+}
+
 static void adjacency()
 {
 	// 0 1/4
@@ -2276,6 +2514,8 @@ void runTests()
 {
 	decodeIndexV0();
 	decodeIndexV1();
+	decodeIndexV1More();
+	decodeIndexV1ThreeEdges();
 	decodeIndex16();
 	encodeIndexMemorySafe();
 	decodeIndexMemorySafe();
@@ -2296,6 +2536,8 @@ void runTests()
 	encodeIndexSequenceEmpty();
 
 	decodeVertexV0();
+	decodeVertexV0More();
+	decodeVertexV0Mode2();
 	decodeVertexV1();
 	decodeVertexV1Custom();
 
@@ -2339,8 +2581,13 @@ void runTests()
 	meshletsSparse();
 	meshletsFlex();
 	meshletsMax();
+	meshletsSpatial();
+	meshletsSpatialDeep();
 
 	partitionBasic();
+	partitionSpatial();
+
+	remapCustom();
 
 	customAllocator();
 
@@ -2366,6 +2613,7 @@ void runTests()
 	simplifyDebug();
 	simplifyPrune();
 	simplifyPruneCleanup();
+	simplifyPruneFunc();
 
 	adjacency();
 	tessellation();
