@@ -66,6 +66,15 @@ static void transformNormal(float* res, const float* ptr, const float* transform
 	res[2] = z * s;
 }
 
+static Stream* getStream(Mesh& mesh, cgltf_attribute_type type, int index = 0)
+{
+	for (size_t i = 0; i < mesh.streams.size(); ++i)
+		if (mesh.streams[i].type == type && mesh.streams[i].index == index)
+			return &mesh.streams[i];
+
+	return NULL;
+}
+
 // assumes mesh & target are structurally identical
 static void transformMesh(Mesh& target, const Mesh& mesh, const cgltf_node* node)
 {
@@ -722,15 +731,6 @@ static void filterTriangles(Mesh& mesh)
 	mesh.indices.resize(write);
 }
 
-static Stream* getStream(Mesh& mesh, cgltf_attribute_type type, int index = 0)
-{
-	for (size_t i = 0; i < mesh.streams.size(); ++i)
-		if (mesh.streams[i].type == type && mesh.streams[i].index == index)
-			return &mesh.streams[i];
-
-	return NULL;
-}
-
 static void simplifyAttributes(std::vector<float>& attrs, float* attrw, size_t stride, Mesh& mesh)
 {
 	assert(stride >= 6); // normal + color
@@ -1199,62 +1199,4 @@ void debugSimplify(const Mesh& source, Mesh& kinds, Mesh& loops, float ratio, fl
 	kinds.streams.push_back(colors);
 	loops.streams.push_back(colors);
 }
-
-void debugMeshlets(const Mesh& source, Mesh& meshlets, int max_vertices)
-{
-	Mesh mesh = source;
-	assert(mesh.type == cgltf_primitive_type_triangles);
-
-	reindexMesh(mesh, /* quantize_tbn= */ true);
-
-	const Stream* positions = getStream(mesh, cgltf_attribute_type_position);
-	assert(positions);
-
-	const float cone_weight = 0.f;
-
-	size_t max_triangles = (max_vertices * 2 + 3) & ~3;
-	size_t max_meshlets = meshopt_buildMeshletsBound(mesh.indices.size(), max_vertices, max_triangles);
-
-	std::vector<meshopt_Meshlet> ml(max_meshlets);
-	std::vector<unsigned int> mlv(max_meshlets * max_vertices);
-	std::vector<unsigned char> mlt(max_meshlets * max_triangles * 3);
-	ml.resize(meshopt_buildMeshlets(&ml[0], &mlv[0], &mlt[0], &mesh.indices[0], mesh.indices.size(), positions->data[0].f, positions->data.size(), sizeof(Attr), max_vertices, max_triangles, cone_weight));
-
-	// generate meshlet meshes, using unique colors
-	meshlets.nodes = mesh.nodes;
-
-	Stream mv = {cgltf_attribute_type_position};
-	Stream mc = {cgltf_attribute_type_color};
-
-	for (size_t i = 0; i < ml.size(); ++i)
-	{
-		const meshopt_Meshlet& m = ml[i];
-
-		unsigned int h = unsigned(i);
-		h ^= h >> 13;
-		h *= 0x5bd1e995;
-		h ^= h >> 15;
-
-		Attr c = {{float(h & 0xff) / 255.f, float((h >> 8) & 0xff) / 255.f, float((h >> 16) & 0xff) / 255.f, 1.f}};
-
-		unsigned int offset = unsigned(mv.data.size());
-
-		for (size_t j = 0; j < m.vertex_count; ++j)
-		{
-			mv.data.push_back(positions->data[mlv[m.vertex_offset + j]]);
-			mc.data.push_back(c);
-		}
-
-		for (size_t j = 0; j < m.triangle_count; ++j)
-		{
-			meshlets.indices.push_back(offset + mlt[m.triangle_offset + j * 3 + 0]);
-			meshlets.indices.push_back(offset + mlt[m.triangle_offset + j * 3 + 1]);
-			meshlets.indices.push_back(offset + mlt[m.triangle_offset + j * 3 + 2]);
-		}
-	}
-
-	meshlets.type = cgltf_primitive_type_triangles;
-	meshlets.streams.push_back(mv);
-	meshlets.streams.push_back(mc);
-}
 #endif
